# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

def in_contrast_or_propagation_scope() -> bool: ...
def has_funchook() -> bool: ...
def get_current_scope() -> tuple: ...
def set_exact_scope(tuple) -> None: ...
