# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Callable, ContextManager

def contrast_scope() -> ContextManager[None]: ...
def propagation_scope() -> ContextManager[None]: ...
def trigger_scope() -> ContextManager[None]: ...
def with_contrast_scope(function: Callable) -> Callable: ...
def in_contrast_scope() -> bool: ...
def in_propagation_scope() -> bool: ...
def in_trigger_scope() -> bool: ...
def enter_contrast_scope() -> None: ...
def exit_contrast_scope() -> None: ...
def enter_propagation_scope() -> None: ...
def exit_propagation_scope() -> None: ...
def enter_trigger_scope() -> None: ...
def exit_trigger_scope() -> None: ...
