# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

DEFAULT_LOG_PATH = "contrast-agent.log"
DEFAULT_SECURITY_LOG_PATH = "security.log"
DEFAULT_SECURITY_LOG_LEVEL = "INFO"
DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_PROGNAME = "Contrast Agent"
SECURITY_LOGGER_NAME = "contrast-security-logger"
LOGGER_NAME = "contrast"
